import * as appsync from "@aws-cdk/aws-appsync";
import * as dynamodb from "@aws-cdk/aws-dynamodb";
import * as cdk from "@aws-cdk/core";
import * as path from "node:path";

export class AppsyncStack extends cdk.Stack {
  constructor(scope: cdk.Construct, id: string, props?: cdk.StackProps) {
    super(scope, id, props);

    const api = new appsync.GraphqlApi(this, "Api", {
      name: "blog-api",
      schema: appsync.Schema.fromAsset(
        path.join(__dirname, "../schema.graphql")
      ),
      authorizationConfig: {
        defaultAuthorization: {
          authorizationType: appsync.AuthorizationType.API_KEY,
          apiKeyConfig: {
            expires: cdk.Expiration.after(cdk.Duration.days(365)),
          },
        },
      },
    });

    new cdk.CfnOutput(this, "GraphQLAPIURL", {
      value: api.graphqlUrl,
    });
    new cdk.CfnOutput(this, "GraphQLAPIKey", {
      value: api.apiKey || "",
    });
    new cdk.CfnOutput(this, "Stack Region", {
      value: this.region,
    });

    const blogTable = new dynamodb.Table(this, "BlogTable", {
      billingMode: dynamodb.BillingMode.PAY_PER_REQUEST,
      partitionKey: {
        name: "id",
        type: dynamodb.AttributeType.STRING,
      },
    });
    const blogDataSource = api.addDynamoDbDataSource(
      "BlogDataSource",
      blogTable
    );
    const postTable = new dynamodb.Table(this, "PostTable", {
      billingMode: dynamodb.BillingMode.PAY_PER_REQUEST,
      partitionKey: {
        name: "id",
        type: dynamodb.AttributeType.STRING,
      },
    });
    postTable.addGlobalSecondaryIndex({
      indexName: "PostedAtIndex",
      partitionKey: {
        name: "blogId",
        type: dynamodb.AttributeType.STRING,
      },
      sortKey: {
        name: "postedAt",
        type: dynamodb.AttributeType.STRING,
      },
    });
    const postDataSource = api.addDynamoDbDataSource(
      "PostDataSource",
      postTable
    );

    blogDataSource.createResolver({
      typeName: "Mutation",
      fieldName: "createBlog",
      requestMappingTemplate: createNodeRequest("Blog"),
      responseMappingTemplate: appsync.MappingTemplate.dynamoDbResultItem(),
    });
    blogDataSource.createResolver({
      typeName: "Mutation",
      fieldName: "updateBlog",
      requestMappingTemplate: appsync.MappingTemplate.dynamoDbPutItem(
        appsync.PrimaryKey.partition("id").is("id"),
        appsync.Values.projecting("input")
      ),
      responseMappingTemplate: appsync.MappingTemplate.dynamoDbResultItem(),
    });
    blogDataSource.createResolver({
      typeName: "Mutation",
      fieldName: "deleteBlog",
      requestMappingTemplate: appsync.MappingTemplate.dynamoDbDeleteItem(
        "id",
        "id"
      ),
      responseMappingTemplate: deleteNodeResponse(),
    });

    postDataSource.createResolver({
      typeName: "Mutation",
      fieldName: "createPost",
      requestMappingTemplate: createNodeRequest("Post", [
        new appsync.Assign("blogId", "$ctx.args.blogId"),
        new appsync.Assign("postedAt", "$util.time.nowISO8601()"),
      ]),
      responseMappingTemplate: appsync.MappingTemplate.dynamoDbResultItem(),
    });
    blogDataSource.createResolver({
      typeName: "Mutation",
      fieldName: "updatePost",
      requestMappingTemplate: appsync.MappingTemplate.dynamoDbPutItem(
        appsync.PrimaryKey.partition("id").is("id"),
        appsync.Values.projecting("input")
      ),
      responseMappingTemplate: appsync.MappingTemplate.dynamoDbResultItem(),
    });
    blogDataSource.createResolver({
      typeName: "Mutation",
      fieldName: "deletePost",
      requestMappingTemplate: appsync.MappingTemplate.dynamoDbDeleteItem(
        "id",
        "id"
      ),
      responseMappingTemplate: deleteNodeResponse(),
    });

    api.createResolver({
      typeName: "Query",
      fieldName: "node",
      requestMappingTemplate: appsync.MappingTemplate.fromString(`
$util.qr($ctx.stash.put("type", $util.base64Decode($ctx.args.id).split(":")[0]))
{}`),
      pipelineConfig: [
        blogDataSource.createFunction({
          name: "blogNode",
          requestMappingTemplate: appsync.MappingTemplate.fromString(`
#if($ctx.stash.type == "Blog")
{
  "version": "2018-05-29",
  "operation": "GetItem",
  "key" : {
    "id" : $util.dynamodb.toDynamoDBJson($ctx.args.id)
  },
}
#else
  #return($ctx.prev.result)
#end`),
          responseMappingTemplate: resultWithTypeNameResponse("Blog"),
        }),
        postDataSource.createFunction({
          name: "postNode",
          requestMappingTemplate: appsync.MappingTemplate.fromString(`
#if($ctx.stash.type == "Post")
{
  "version": "2018-05-29",
  "operation": "GetItem",
  "key" : {
    "id" : $util.dynamodb.toDynamoDBJson($ctx.args.id)
  },
}
#else
  #return($ctx.prev.result)
#end`),
          responseMappingTemplate: resultWithTypeNameResponse("Post"),
        }),
      ],
      responseMappingTemplate: appsync.MappingTemplate.dynamoDbResultItem(),
    });

    blogDataSource.createResolver({
      typeName: "Query",
      fieldName: "blogs",
      requestMappingTemplate: appsync.MappingTemplate.fromString(`
{
  "version" : "2018-05-29",
  "operation" : "Scan",
  "nextToken" : #if($ctx.args.after) "$ctx.args.after" #else null #end,
  "limit" : $ctx.args.first
}`),
      responseMappingTemplate: connectionResponse(),
    });

    postDataSource.createResolver({
      typeName: "Blog",
      fieldName: "posts",
      requestMappingTemplate: appsync.MappingTemplate.fromString(`
{
  "version" : "2018-05-29",
  "operation" : "Query",
  "query" : {
    "expression" : "blogId = :blogId",
    "expressionValues" : {
      ":blogId" : $util.dynamodb.toDynamoDBJson($ctx.source.id)
    }
  },
  "index" : "PostedAtIndex",
  "scanIndexForward" : false,
  "nextToken" : #if($ctx.args.after) "$ctx.args.after" #else null #end,
  "limit" : $ctx.args.first
}`),
      responseMappingTemplate: connectionResponse(),
    });
  }
}

function createNodeRequest(
  typeName: string,
  assign?: appsync.Assign[]
): appsync.MappingTemplate {
  const attr = new appsync.AttributeValues("$ctx.args.input", assign);
  return appsync.MappingTemplate.fromString(`
${attr.renderVariables()}
#set($id = $util.autoId())
{
  "version": "2018-05-29",
  "operation": "PutItem",
  "key" : {
    "id" : $util.dynamodb.toDynamoDBJson($util.base64Encode("${typeName}:\${id}"))
  },
  ${attr.renderTemplate()}
}`);
}

function deleteNodeResponse(): appsync.MappingTemplate {
  return appsync.MappingTemplate.fromString(`
$util.toJson($ctx.result.id)`);
}

function resultWithTypeNameResponse(typeName: string): appsync.MappingTemplate {
  return appsync.MappingTemplate.fromString(`
#set($result = $ctx.result)
$util.qr($result.put("__typename", "${typeName}"))
$util.toJson($result)`);
}

function connectionResponse(): appsync.MappingTemplate {
  return appsync.MappingTemplate.fromString(`
#set($edges = [])
#foreach($item in $ctx.result.items)
  $util.qr($edges.add({"node": $item, "cursor": ""}))
#end
{
  "edges": $util.toJson($edges),
  "pageInfo": {
    "hasPreviousPage": false,
    "hasNextPage": #if($ctx.result.nextToken) true #else false #end,
    "endCursor": #if($ctx.result.nextToken) "$context.result.nextToken" #else null #end
  }
}`);
}
